/** -*- linux-c -*- ***********************************************************
 * Linux PPP over L2TP (PPPoX/PPPoL2TP) Sockets
 *
 * PPPoX    --- Generic PPP encapsulation socket family
 * PPPoL2TP --- PPP over L2TP (RFC 2661)
 *
 *
 * Version:    0.0.1
 *
 * 251003 :     Copied from pppoe.c version 0.6.9.
 &
 * Author:	Martijn van Oosterhout <kleptog@svana.org>
 * Contributors:
 *              Michal Ostrowski <mostrows@speakeasy.net>
 * 		Arnaldo Carvalho de Melo <acme@xconectiva.com.br>
 *		David S. Miller (davem@redhat.com)
 *
 * License:
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 */

/**
 * How the kernel structures are arranged:
 *
 * struct sock *server_sock->user_data = struct pppol2tp_sock
 * struck sock *PPPoL2TP_sock->user_data = struct pppol2tp_call
 
 * struct pppol2tp_sock->sock_ops = old server_sock ops
 * struct pppol2tp_sock->hash = hash table of struct pppol2tp_calls
 
 * struct pppol2tp_call->sock = the PPPoL2TP sock
 * struct pppol2tp_call->server_sock = the server socket
 
 */

#include <linux/string.h>
#include <linux/module.h>

#include <asm/uaccess.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/errno.h>

#include <linux/netdevice.h>
#include <linux/net.h>
#include <linux/inetdevice.h>
//#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/init.h>
//#include <linux/if_ether.h>
#include <linux/udp.h>
#include <linux/if_pppox.h>
#include <net/sock.h>
#include <linux/ppp_channel.h>
#include <linux/ppp_defs.h>
#include <linux/if_ppp.h>
#include <linux/if_pppvar.h>
//#include <linux/notifier.h>
#include <linux/file.h>
#include <linux/proc_fs.h>

#define MYSTATIC 

#define L2TP_HDRFLAG_T     0x8000
#define L2TP_HDRFLAG_L     0x4000
#define L2TP_HDRFLAG_S     0x0800
#define L2TP_HDRFLAG_O     0x0200
#define L2TP_HDRFLAG_P     0x0100

#define L2TP_HDR_VER_MASK  0x000F
#define L2TP_HDR_VER       0x0002

/* Just some random numbers */
#define L2TP_MAGIC          0x42114DDA
#define L2TP_SESSION_MAGIC  0x0C04EB7D

MYSTATIC int __attribute__((unused)) pppol2tp_debug = 7;
#define PPPOL2TP_HASH_BITS 4
#define PPPOL2TP_HASH_SIZE (1<<PPPOL2TP_HASH_BITS)

/* FIXME: Shouldn't need this but anyway */
extern struct proto udp_prot;

/* Describes a session. It is the user_data field in the PPPOX socket. Contains
 * just enough information to determine incoming packets and transmit
 * outgoing ones. */
struct pppol2tp_session
{
	int magic;
	
	struct sock *sock;         /* Pointer to the PPPoX socket */
	struct sock *server_sock;  /* Pointer to the UDP socket */
	
	struct pppol2tp_addr tunnel_addr;  /* Description of tunnel */
	
	struct pppol2tp_session *next;   /* Next in this hash chain */
};

/* The user_data field of the server's UDP socket. It contains info to track
 * all the associated sessions so incoming packets can be sorted out */
struct pppol2tp_sock
{
	int magic;  /* Should be L2TP_MAGIC */
	
	struct proto old_proto,       /* original proto */
	             l2tp_proto;      /* L2TP proto */
	struct pppol2tp_session *hash[ PPPOL2TP_HASH_SIZE ];
	
	void (*old_data_ready)( struct sock *, int );

	struct sock *sock;            /* Parent socket */	
	struct pppol2tp_sock *next;   /* Keep a list of all open prepared sockets */
};

/* This isn't the real L2TP header, just the one we use to send data frames
 * out. The real format has fields that move all over the place and can't be
 * described by a structure */

struct pppol2tp_hdr
{
	u16     flags;     /* Must be set to htons(L2TP_HDR_VER | L2TP_HDRFLAG_L) */
	u16     tunnel;
	u16     session;
	u16     offset;    /* Set to zero */
};

MYSTATIC int pppol2tp_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg);
MYSTATIC int pppol2tp_xmit(struct ppp_channel *chan, struct sk_buff *skb);
MYSTATIC int __pppol2tp_xmit(struct sock *sk, struct sk_buff *skb);

MYSTATIC struct ppp_channel_ops pppol2tp_chan_ops = { pppol2tp_xmit , NULL };
MYSTATIC struct proto_ops pppol2tp_ops;
MYSTATIC struct pppol2tp_sock *pppol2tp_socket_list;

#if 0
#define CHECKPTR(x,y) { if (!(x) && pppol2tp_debug &7 ){ printk(KERN_CRIT "PPPoL2TP Invalid pointer : %s , %p\n",#x,(x)); error=-EINVAL; goto y; }}
#define DEBUG(s,args...) if( pppol2tp_debug & (s) ) printk(KERN_CRIT args );
#else
#define CHECKPTR(x,y) do {} while (0);
#define DEBUG(s,args...) do { } while (0);
#endif

#define SOCK_2_SESSION( sock, session, err, errval, label ) \
	session = (struct pppol2tp_session *)((sock)->user_data);  \
	if( !session || session->magic != L2TP_SESSION_MAGIC ) {         \
		printk( KERN_CRIT "%s: %s:%d: BAD SESSION MAGIC (" #sock "=%p)\n", __FUNCTION__, __FILE__, __LINE__, sock ); \
		err = errval; \
		goto label; \
	}
	
#define SOCK_2_SERVER( sock, server, err, errval, label ) \
	server = (struct pppol2tp_sock *)((sock)->user_data);  \
	if( !server || server->magic != L2TP_MAGIC ) {         \
		printk( KERN_CRIT "%s: %s:%d: BAD MAGIC (" #sock "=%p)\n", __FUNCTION__, __FILE__, __LINE__, sock ); \
		err = errval; \
		goto label; \
	}

#define ENTER_FUNCTION   printk( KERN_CRIT "--> %s\n", __FUNCTION__ )
#define EXIT_FUNCTION    printk( KERN_CRIT "<-- %s\n", __FUNCTION__ )

/* Copied from socket.c */
extern __inline__ void sockfd_put(struct socket *sock)
{
	fput(sock->file);
}

MYSTATIC rwlock_t pppol2tp_hash_lock = RW_LOCK_UNLOCKED;

MYSTATIC inline int cmp_addr(struct pppol2tp_addr *a, u16 tunnel, u16 session)
{
	return (a->s_tunnel == tunnel && a->s_session == session);
}

MYSTATIC int hash_item(u16 tunnel, u16 session)
{
	int i, temp, hash;

	hash = 0;	
	temp = tunnel ^ session;

	for (i = 0; i <= sizeof(u16)/PPPOL2TP_HASH_BITS ; ++i) {
		hash ^= temp;
		temp >>= PPPOL2TP_HASH_BITS;
	}
	return hash & ( PPPOL2TP_HASH_SIZE - 1 );
}

/**********************************************************************
 *
 *  Set/get/delete/rehash items  (internal versions)
 *
 **********************************************************************/
MYSTATIC struct pppol2tp_session *__get_item(struct pppol2tp_sock *sock, u16 tunnel, u16 session)
{
	int hash = hash_item(tunnel, session);
	struct pppol2tp_session *ret;

	ret = sock->hash[ hash ];

	while (ret && !cmp_addr(&ret->tunnel_addr, tunnel, session))
		ret = ret->next;

	return ret;
}

MYSTATIC int __set_item(struct pppol2tp_sock *sock, struct pppol2tp_session *po)
{
	struct pppol2tp_addr *addr = &po->tunnel_addr;
	int hash = hash_item(addr->s_tunnel, addr->s_session);
	struct pppol2tp_session *ret;

	ret = sock->hash[hash];
	while (ret) {
		if (cmp_addr(&ret->tunnel_addr, addr->s_tunnel, addr->s_session))
			return -EALREADY;

		ret = ret->next;
	}

	if (!ret) {
		po->next = sock->hash[hash];
		sock->hash[hash] = po;
	}

	return 0;
}

MYSTATIC struct pppol2tp_session *__delete_item(struct pppol2tp_sock *sock, u16 tunnel, u16 session)
{
	int hash = hash_item(tunnel, session);
	struct pppol2tp_session *ret, **src;

	ret = sock->hash[hash];
	src = &sock->hash[hash];

	while (ret) {
		if (cmp_addr(&ret->tunnel_addr, tunnel, session)) {
			*src = ret->next;
			break;
		}

		src = &ret->next;
		ret = ret->next;
	}

	return ret;
}

/**********************************************************************
 *
 *  Set/get/delete/rehash items
 *
 **********************************************************************/
MYSTATIC inline struct pppol2tp_session *get_item(struct pppol2tp_sock *sock, u16 tunnel, u16 session)
{
	struct pppol2tp_session *po;
	
	ENTER_FUNCTION;

	read_lock_bh(&pppol2tp_hash_lock);
	po = __get_item(sock, tunnel, session);
	if (po)
		sock_hold(po->sock);
	read_unlock_bh(&pppol2tp_hash_lock);

	EXIT_FUNCTION;
	
	return po;
}

#if 0
MYSTATIC inline struct pppol2tp_session *get_item_by_addr()
{
	return get_item(sp->sa_addr.pppol2tp.sid, sp->sa_addr.pppol2tp.remote);
}
#endif

MYSTATIC inline int set_item(struct pppol2tp_sock *sock, struct pppol2tp_session *po)
{
	int i;
	
	ENTER_FUNCTION;
	
	printk( KERN_ERR "set_item( sock=%p, po=%p ) = ", sock, po );

	if (!po)
		return -EINVAL;

	write_lock_bh(&pppol2tp_hash_lock);
	i = __set_item(sock, po);
	write_unlock_bh(&pppol2tp_hash_lock);

	printk( "%d\n", i );
	
	EXIT_FUNCTION;
	return i;
}

MYSTATIC inline struct pppol2tp_session *delete_item(struct pppol2tp_sock *sock, u16 tunnel, u16 session)
{
	struct pppol2tp_session *ret;
	ENTER_FUNCTION;
	write_lock_bh(&pppol2tp_hash_lock);
	ret = __delete_item(sock, tunnel, session);
	write_unlock_bh(&pppol2tp_hash_lock);
	EXIT_FUNCTION;
	return ret;
}

/***************************************************************************
 *
 *  When the server UDP socket is closed, all the attached sockets need to
 *  go too. This handles that.
 *
 **************************************************************************/

MYSTATIC void pppol2tp_server_closeall(struct pppol2tp_sock *sock)
{
	int hash;

	if (sock == NULL)
		BUG();

	ENTER_FUNCTION;
	
	read_lock_bh(&pppol2tp_hash_lock);
	for (hash = 0; hash < PPPOL2TP_HASH_SIZE; hash++) {
		struct pppol2tp_session *po = sock->hash[hash];

		while (po != NULL) {
			struct sock *sk = po->sock;

			sock_hold(sk);

			/* We hold a reference to SK, now drop the
			 * hash table lock so that we may attempt
			 * to lock the socket (which can sleep).
			 */
			read_unlock_bh(&pppol2tp_hash_lock);

			lock_sock(sk);

			if (sk->state & (PPPOX_CONNECTED | PPPOX_BOUND)) {
				pppox_unbind_sock(sk);
				sk->state = PPPOX_DEAD;
				sk->state_change(sk);
			}

			release_sock(sk);

			sock_put(sk);

			read_lock_bh(&pppol2tp_hash_lock);

			po = po->next;
		}
		sock->hash[hash] = NULL;
	}
	read_unlock_bh(&pppol2tp_hash_lock);
	EXIT_FUNCTION;
}

/*************************************************************************
 * Server UDP socket destruct hook
 ************************************************************************/
MYSTATIC void pppol2tp_server_destruct( struct sock *sk )
{
	struct pppol2tp_sock *l2tp_sock;
	struct pppol2tp_sock **ptr;
	int error = 0;
	ENTER_FUNCTION;
	
	SOCK_2_SERVER( sk, l2tp_sock, error, -EBADF, end );
	
	pppol2tp_server_closeall( l2tp_sock );

	// Remove from socket list
	for( ptr = &pppol2tp_socket_list; *ptr; ptr = &(*ptr)->next )
	{
		if( *ptr == l2tp_sock )
		{
			*ptr = l2tp_sock->next;
			break;
		}
	}
	printk( KERN_ERR "pppol2tp_server_destruct: MOD_DEC_USE_COUNT\n" );
	MOD_DEC_USE_COUNT;
end:
	EXIT_FUNCTION;
	return;
}

/************************************************************************
 *
 * Do the real work of receiving an L2TP data frame.
 *
 ***********************************************************************/
int pppol2tp_recv_core(struct sock *sock, struct sk_buff *skb)
{
	struct pppol2tp_session *l2tp_session = NULL;
	int error = 0;
	
	/* TODO: Check magic number */
	struct pppol2tp_sock *l2tp_sock;
	struct sock *session_sock = NULL;
	
	unsigned char *ptr;
	u16 hdrflags;
	u16 tunnel, session;
	int length;

	ENTER_FUNCTION;
	
	ptr = skb->data + sizeof( struct udphdr );
	hdrflags = ntohs( *(u16*)ptr );
	SOCK_2_SERVER( sock, l2tp_sock, error, 1, end );

	printk( KERN_ERR "pppol2tp_recv_core(sock=%p, skb=%p) l2tp_sock=%p\n" KERN_ERR, sock, skb, l2tp_sock );

	for( length = 0; length < 16; length++ )
		printk( " %02X", ptr[length] );
	printk( "\n" );
	
	length=ntohs(skb->h.uh->len)-sizeof(struct udphdr);
	
	if( length < 12 )  /* Short packet? */
	{
		printk( "pppol2tp_recv_core: short packet (len=%d)\n", length );
		return 1;
	}
	
	if( hdrflags & L2TP_HDRFLAG_T )   /* If type is control packet */
	{
		printk( "pppol2tp_recv_core: control packet\n" );
		return 1;
	}

	ptr+=2;  /* Skip flags */
	
	if ( hdrflags & L2TP_HDRFLAG_L )  /* If length is present, skip it */
		ptr+=2;
		
	tunnel  = ntohs( *(u16*)ptr );   /* Extract tunnel and session ID */
	ptr += 2;
	session = ntohs( *(u16*)ptr );
	ptr += 2;

	l2tp_session = get_item(l2tp_sock, tunnel, session);

	if (!l2tp_session) {   /* Not found? Pass to userspace to deal with */
		printk( "pppol2tp_recv_core: no socket found (%04X:%04X)\n", tunnel, session );
		return 1;
	}

	if ( hdrflags & L2TP_HDRFLAG_S )  /* If sequence bit set, skip them */
		ptr += 4;
		
	if ( hdrflags & L2TP_HDRFLAG_O )  /* If offset bit set, skip it */
		ptr += 2 + ntohs( *(u16*)ptr );  
		
	session_sock = l2tp_session->sock;
	skb_pull(skb, ptr - skb->data);
	skb_unlink( skb );
	if (session_sock->state & PPPOX_BOUND)
	{
		printk( "pppol2tp_recv_core: passing to ppp_input\n" );
		ppp_input(&session_sock->protinfo.pppox->chan, skb);
	}
	else
	{
		printk( "pppol2tp_recv_core: socket not bound\n" );
		/* Not bound??? Queue it now */
		sock_queue_rcv_skb( session_sock, skb );
	}
	sock_put( l2tp_session->sock );
	EXIT_FUNCTION;
	return 0;
end:
	EXIT_FUNCTION;
	return 1;
}

/************************************************************************
 *
 * Receive message. This is the recvmsg for the PPPoL2TP socket.
 *
 ***********************************************************************/
MYSTATIC int pppol2tp_recvmsg(struct socket *sock, struct msghdr *msg, int len,
		 int flags, struct scm_cookie *scm)
{
	int err = 0;
  	struct sk_buff *skb = NULL;
  	struct sock *sk = sock->sk;

	ENTER_FUNCTION;

	err = -EIO;
	if( sock->state & PPPOX_BOUND )
		goto error;
			
	msg->msg_namelen = 0;
	
	skb=skb_recv_datagram(sk,flags & ~MSG_DONTWAIT,
	                         flags & MSG_DONTWAIT,&err);
        if (skb) {
                err = memcpy_toiovec(msg->msg_iov, (unsigned char *) skb->data, skb->len);
                if (err < 0)
                        goto do_skb_free;
                err = skb->len;
        }
do_skb_free:
	if( skb )
		kfree_skb( skb );
error:
	EXIT_FUNCTION;
	return err;
}

/***********************************************************************
 *
 * Really kill the socket. (Called from sock_put if refcnt == 0.)
 *
 **********************************************************************/
void pppol2tp_sock_destruct(struct sock *sk)
{
	ENTER_FUNCTION;
	
	if (sk->protinfo.destruct_hook)
		kfree(sk->protinfo.destruct_hook);
	printk( KERN_ERR "pppol2tp_sock_destruct: MOD_DEC_USE_COUNT\n" );
	MOD_DEC_USE_COUNT;
	EXIT_FUNCTION;
}


/***********************************************************************
 *
 * Initialize a new struct sock. This is for the session sockets.
 *
 **********************************************************************/
MYSTATIC int pppol2tp_create(struct socket *sock)
{
	int error = 0;
	struct sock *sk;

	ENTER_FUNCTION;
	MOD_INC_USE_COUNT;
	printk( KERN_ERR "pppol2tp_create: MOD_INC_USE_COUNT\n" );
	printk( KERN_ERR "pppol2tp_create( sock=%p )\n", sock );

	sk = sk_alloc(PF_PPPOX, GFP_KERNEL, 1);
	if (!sk)
		return -ENOMEM;

	sock_init_data(sock, sk);

	sock->state = SS_UNCONNECTED;
	sock->ops   = &pppol2tp_ops;

	sk->protocol = PX_PROTO_OL2TP;
	sk->family = PF_PPPOX;

//	sk->backlog_rcv = pppol2tp_recv_core;

	sk->next = NULL;
	sk->pprev = NULL;
	sk->state = PPPOX_NONE;
	sk->type = SOCK_STREAM;
	sk->destruct = pppol2tp_sock_destruct;

	sk->protinfo.pppox = kmalloc(sizeof(struct pppox_opt), GFP_KERNEL);
	if (!sk->protinfo.pppox) {
		error = -ENOMEM;
		goto free_sk;
	}

	memset((void *) sk->protinfo.pppox, 0, sizeof(struct pppox_opt));
	sk->protinfo.pppox->sk = sk;

	/* Delete the protinfo when it is time to do so. */
	/* Since this is a union, isn't this assigning to itself? */
	sk->protinfo.destruct_hook = sk->protinfo.pppox;
	sock->sk = sk;

	EXIT_FUNCTION;
	return 0;

free_sk:
	sk_free(sk);
	EXIT_FUNCTION;
	return error;
}

/****************************************************************
 * Called when the PPPoX socket is closed
 ***************************************************************/
int pppol2tp_release(struct socket *sock)
{
	struct sock *sk = sock->sk;
	struct pppox_opt *po;
	struct pppol2tp_session *session_sock;
	struct pppol2tp_sock *server_sock;
	int error = 0;
	ENTER_FUNCTION;

	if (!sk)
		return 0;

	if (sk->dead != 0)
		return -EBADF;

	po = sk->protinfo.pppox;
	if( sk->user_data )           /* Was this socket actually connected? */
	{
		SOCK_2_SESSION( sk, session_sock, error, -EBADF, end );
		SOCK_2_SERVER( session_sock->server_sock, server_sock, error, -EBADF, end );
		/* Delete the session socket from the hash */
		delete_item(server_sock, session_sock->tunnel_addr.s_tunnel, session_sock->tunnel_addr.s_session);
	}
	pppox_unbind_sock(sk);

	/* Signal the death of the socket. */
	sk->state = PPPOX_DEAD;

	sock_orphan(sk);
	sock->sk = NULL;

	/* FIXME: If I uncomment this I get a kernel panic */
	skb_queue_purge(&sk->receive_queue);
	printk( KERN_ERR "pppol2tp_release: Calling sock_put; refcnt=%d\n", sk->refcnt.counter );
	sock_put(sk);

end:
	EXIT_FUNCTION;
	return error;
}

/********************************************************************
 * The data_ready hook on the UDP socket. Scan the incoming packet list for
 * packets to process
 *******************************************************************/
MYSTATIC void pppol2tp_data_ready( struct sock *sk, int len )
{
	int err;
	struct pppol2tp_sock *server_sock;
	struct sk_buff *skb;
	int processed = 0;
	
	ENTER_FUNCTION;
	SOCK_2_SERVER( sk, server_sock, err, -EBADF, end );
	
	printk( KERN_ERR "pppol2tp_data_ready: Received %d bytes on socket %p\n", len, sk );
	
	/* FIXME: Do we need to lock the socket here? */
	skb_queue_walk( &sk->receive_queue, skb )
	{
		if( pppol2tp_recv_core( sk, skb ) )   /* Pass to userspace */
		{
			printk( KERN_ERR "pppol2tp_data_ready: Packet passed to userspace\n" );
		}
		else
		{
			printk( KERN_ERR "pppol2tp_data_ready: Packet accepted\n" );
			processed = 1;
			/* If the packet has been accepted it is the
			 * responsibility of the receiver (either socket or ppp
			 * device) to dispose it) */
			 
			/* Also, since recv_core has requeued the packet
			 * elsewhere, it's not safe to continue this loop,
			 * so we break */
			break;
		}
	}
	if( !processed )
	{
		printk( KERN_ERR "pppol2tp_data_ready: Calling old old_data_ready\n" );
		server_sock->old_data_ready( sk, len );
	}
end:
	EXIT_FUNCTION;
	return;
}

/******************************************************************************
 * Prepare a server UDP socket to have PPPoX sockets attached to it
 *****************************************************************************/
MYSTATIC struct sock *pppol2tp_prepare_socket( int fd, int *error )
{
	int err;
#if 0
	struct file *file;
	struct inode *inode;
#endif
	struct socket *sock = NULL;
	struct sock *sk;
	struct pppol2tp_sock *l2tp_server;
    	ENTER_FUNCTION;
    	
    	printk( KERN_ERR "pppol2tp_prepare_socket( fd=%d, error=%p(%d) )\n", fd, error, *error );
    	
	err = -EBADF;
	printk( KERN_ERR "pppol2tp_prepare_socket: sockfd_lookup( fd=%d, &err=%p ) => ", fd, &err );
	sock = sockfd_lookup(fd,&err);
	printk( "%p (err=%d)\n", sock, err );
	if ( !sock )
		goto err;
#if 0
	if (sock->file != file) {
		printk(KERN_ERR "pppol2tp: socki_lookup: socket file changed!\n");
		sock->file = file;
	}
	fput(file);
#endif
	err = -ESOCKTNOSUPPORT;
	if (sock->type!=SOCK_DGRAM)
	{
		printk( KERN_ERR "pppol2tp_prepare_socket: fd %d wrong type, got %d, expected %d\n", fd, sock->type, SOCK_DGRAM );
		/* Because for some reason this happens all the time */
//		goto err;
	}
	err = -EAFNOSUPPORT;
	if (sock->ops->family!=AF_INET)
	{
		printk( KERN_ERR "pppol2tp_prepare_socket: fd %d wrong family, got %d, expected %d\n", fd, sock->ops->family, AF_INET );
		goto err;
	}

	err = -ENOTCONN;
	sk=sock->sk;
	printk( KERN_ERR "pppol2tp_prepare_socket: sk->prot=%p\n", sk->prot );
#if 0
/* We don't require a connected socket */
	if (sk->state!=TCP_ESTABLISHED)
	{
		printk( KERN_ERR "pppol2tp_prepare_socket: fd %d wrong state, got %d, expected %d\n", fd, sk->state, TCP_ESTABLISHED );
		goto err;
	}
#endif
	printk( KERN_ERR "pppol2tp_prepare_socket: Checking user_data (%p)\n", sk->user_data );
	
	l2tp_server = (struct pppol2tp_sock *)sk->user_data;
	if( l2tp_server )   /* User-data field already set */
	{
		if( l2tp_server < (struct pppol2tp_sock *)0x10000000 )
		{
			printk( KERN_ERR "pppol2tp_prepare_socket: WARNING: user_data set with apparently invalid value, overwriting (%p)\n", l2tp_server );
		}
		else
		{
			err = -EBUSY;
			if( l2tp_server->magic != L2TP_MAGIC )
				goto err;
			/* This socket has already been prepped */
			EXIT_FUNCTION;
			return l2tp_server->sock;
		}
	}
	/* This socket is available and needs prepping */
	sk->user_data = l2tp_server = kmalloc( sizeof( struct pppol2tp_sock ), GFP_KERNEL );
#if 0	
	printk( KERN_ERR "pppol2tp_prepare_socket: Allocated l2tp_server=%p (sk=%p,sock=%p)\n", l2tp_server, sk, sock );
	
	printk( KERN_ERR "NOTE: sk(%p)->user_data = %p\n", sk, sk->user_data );
#endif
	memset( l2tp_server, 0, sizeof( struct pppol2tp_sock ) );
	
	l2tp_server->magic = L2TP_MAGIC;

	/* Setup the new protocol stuff */
	/* FIXME: WHY IS sk->prot NULL !!! */
//	if( sk->prot )
//	{
		l2tp_server->old_proto = *sk->prot;
		l2tp_server->l2tp_proto = *sk->prot;
//	}
#if 0
	else
	{
		l2tp_server->old_proto = l2tp_server->l2tp_proto = udp_prot;
	}
#endif
	
//	l2tp_server->l2tp_proto.close = &pppol2tp_server_close;
//	l2tp_server->l2tp_proto.recvmsg = &pppol2tp_recvmsg;
	sk->prot = &l2tp_server->l2tp_proto;
	
#if 0
	l2tp_server->old_ops = sock->ops;
	l2tp_server->l2tp_ops = *l2tp_server->old_ops;
	l2tp_server->l2tp_ops.release = &pppol2tp_server_release;
	sock->ops = &l2tp_server->l2tp_ops;
#endif	
	l2tp_server->old_data_ready = sk->data_ready;
	sk->data_ready = &pppol2tp_data_ready;

	printk( KERN_ERR "pppol2tp_prepare_socket: Old destruct value = %p\n", sk->destruct );
	sk->destruct = &pppol2tp_server_destruct;
	
	l2tp_server->next = pppol2tp_socket_list;
	pppol2tp_socket_list = l2tp_server;
	
	l2tp_server->sock = sk;
	
	MOD_INC_USE_COUNT;
	printk( KERN_ERR "pppol2tp_prepare_socket: MOD_INC_USE_COUNT\n" );
	
	if( sock )
		sockfd_put( sock );
	EXIT_FUNCTION;
	return l2tp_server->sock;
err:
	*error = err;
	if( sock )
		sockfd_put( sock );
	EXIT_FUNCTION;
	return NULL;
}

/************************************************************************
 * Attach a PPPoX socket to a server UDP socket 
 ***********************************************************************/
int pppol2tp_connect(struct socket *sock, struct sockaddr *uservaddr,
		  int sockaddr_len, int flags)
{
	struct sock *sk = sock->sk;
	struct sockaddr_pppox *sp = (struct sockaddr_pppox *) uservaddr;
	struct pppox_opt *po = sk->protinfo.pppox;
	struct sock *server_sock = NULL;
	struct pppol2tp_session *l2tp_session;
	struct pppol2tp_sock *l2tp_server;
	int error = 0;

	ENTER_FUNCTION;
	
	printk( KERN_ERR "pppol2tp_connect( sock=%p, uservaddr=%p, sockaddr_len=%d, flag=%d )\n", sock, uservaddr, sockaddr_len, flags );
	
	printk( KERN_ERR "pppol2tp_connect: lock_sock( sk=%p )\n", sk );
	lock_sock(sk);

	error = -EINVAL;
	if (sp->sa_protocol != PX_PROTO_OL2TP)
		goto end;

	/* Check for already bound sockets */
	error = -EBUSY;
	if (sk->state & PPPOX_CONNECTED)
		goto end;

	/* We don't supporting rebinding anyway */		
	if ( sk->user_data )
		goto end; /* socket is already attached */

	/* Don't bind if s_tunnel is 0 */
	error = -EINVAL;
	if (sp->sa_addr.pppol2tp.s_tunnel == 0)
		goto end;

	/* This looks up the server socket and configures it if necessary */
	server_sock = pppol2tp_prepare_socket( sp->sa_addr.pppol2tp.fd, &error );
	if( server_sock == NULL )
		goto end;

	printk( KERN_ERR "NOTE: server_sock(%p)->user_data = %p\n", server_sock, server_sock->user_data );
	l2tp_session = kmalloc( sizeof( struct pppol2tp_session ), GFP_KERNEL );
	memset( l2tp_session, 0, sizeof( struct pppol2tp_session ) );

	l2tp_session->magic = L2TP_SESSION_MAGIC;
	l2tp_session->sock = sk;
	l2tp_session->server_sock = server_sock;
	l2tp_session->tunnel_addr = sp->sa_addr.pppol2tp;
	
	SOCK_2_SERVER( server_sock, l2tp_server, error, -EBADF, end );
	if( set_item( l2tp_server, l2tp_session ) )
	{
		printk( KERN_ERR "pppol2tp_connect: set_item *FAILED*\n" );
		kfree( l2tp_session );
		error = -EALREADY;
		goto end;
	}
	
	sk->user_data = l2tp_session;
	
	/* We don't store any more options in the pppox_opt, everything is
	* in user_data (struct pppol2tp_session) */
	po->sk = sk;

	/* Right now, because we don't have a way to push the incoming skb's
	 * straight through the UDP layer, the only header we need to worry
	 * about is the L2TP header */
	
	po->chan.hdrlen = sizeof(struct pppol2tp_hdr);

	po->chan.private = sk;
	po->chan.ops = &pppol2tp_chan_ops;

	error = ppp_register_channel(&po->chan);
	if (error)
		goto end;

	sk->state = PPPOX_CONNECTED;

 end:
 	/* For some reason the kernel panics on this... */
	release_sock(sk);
	printk( KERN_ERR "NOTE: server_sock(%p)->user_data = %p\n", server_sock, server_sock->user_data );
	printk( KERN_ERR "pppol2tp_connect: should release_sock(%p) here\n", sk );
	EXIT_FUNCTION;
	return error;
}


/************************************************************************
 * Return tunnal address
 ***********************************************************************/
int pppol2tp_getname(struct socket *sock, struct sockaddr *uaddr,
		  int *usockaddr_len, int peer)
{
	int len = sizeof(struct sockaddr_pppox);
	struct sockaddr_pppox sp;
	int error = 0;
	struct pppol2tp_session *l2tp_session;

	ENTER_FUNCTION;
	
	SOCK_2_SESSION( sock->sk, l2tp_session, error, -EBADF, end );
	
	sp.sa_family	= AF_PPPOX;
	sp.sa_protocol	= PX_PROTO_OL2TP;
	memcpy(&sp.sa_addr.pppol2tp, &l2tp_session->tunnel_addr,
	       sizeof(struct pppol2tp_addr));

	memcpy(uaddr, &sp, len);

	*usockaddr_len = len;

end:
	EXIT_FUNCTION;
	return error;
}


/*************************************************************************
 * Handle ioctl()s
 ************************************************************************/
MYSTATIC int pppol2tp_ioctl(struct socket *sock, unsigned int cmd,
		unsigned long arg)
{
	struct sock *sk = sock->sk;
	struct pppox_opt *po;
	int val = 0;
	int err = 0;

	ENTER_FUNCTION;
	
	printk( KERN_ERR "pppol2tp_ioctl( sock=%p, cmd=%#x, arg=%#lx )\n", sock, cmd, arg );
	
	po = sk->protinfo.pppox;
	switch (cmd) {
	case PPPIOCGMRU:
		err = -ENXIO;

		if (!(sk->state & PPPOX_CONNECTED))
			break;

		err = -EFAULT;
		if (put_user(1460,             /* Erm, not really. Got a better idea? */
			     (int *) arg))
			break;
		err = 0;
		break;

	case PPPIOCSMRU:
		err = -ENXIO;
		if (!(sk->state & PPPOX_CONNECTED))
			break;

		err = -EFAULT;
		if (get_user(val,(int *) arg))
			break;

		if (val < 1460)             /* Erm, not really. Got a better idea? */
			err = 0;
		else
			err = -EINVAL;
		break;

	case PPPIOCSFLAGS:
		err = -EFAULT;
		if (get_user(val, (int *) arg))
			break;
		err = 0;
		break;

	default:;
	};

	EXIT_FUNCTION;
	return err;
}

/**************************************************************************
 * This is the sendmsg for the PPPOX pppol2tp_session socket
 *************************************************************************/
int pppol2tp_sendmsg(struct socket *sock, struct msghdr *m,
		  int total_len, struct scm_cookie *scm )
{
	struct sock *sk = sock->sk;
	int error = 0;
	struct pppol2tp_hdr hdr;
	struct msghdr *msg;
	struct pppol2tp_session *l2tp_session;
	struct pppol2tp_sock *l2tp_sock;
	ENTER_FUNCTION;
	if (sk->dead || !(sk->state & PPPOX_CONNECTED)) {
		error = -ENOTCONN;
		goto end;
	}

	/* FIXME: Check magic */
	SOCK_2_SESSION( sk, l2tp_session, error, -EBADF, end );
	SOCK_2_SERVER( l2tp_session->server_sock, l2tp_sock, error, -EBADF, end );

	/* Setup L2TP header */	
	hdr.flags  = htons(L2TP_HDR_VER | L2TP_HDRFLAG_O);
	hdr.tunnel = htons( l2tp_session->tunnel_addr.d_tunnel );
	hdr.session= htons( l2tp_session->tunnel_addr.d_session );
	hdr.offset = 0;

	/* The old code had a size limit here. It's not really an issue
	 * since the upper layers will fragment anyway if necessary */
	
	/* Unfortunatly there doesn't appear to be a way for us to pass an
	 * skb to the UDP layer, we have to pretend to be sending ordinary
	 * data and use sendmsg */
	
	msg = kmalloc( sizeof( struct msghdr ), GFP_KERNEL );
	
	msg->msg_iov = kmalloc( (m->msg_iovlen+1) * sizeof( struct iovec ), GFP_KERNEL );
	msg->msg_iov[0].iov_base = &hdr;
	msg->msg_iov[0].iov_len  = sizeof( hdr );
	memcpy( &msg->msg_iov[1], &m->msg_iov[0], m->msg_iovlen * sizeof( struct iovec ) );
	msg->msg_iovlen = m->msg_iovlen+1;
	
	/* If the user calls sendto() that's just too bad */
	msg->msg_name = &l2tp_session->tunnel_addr.addr;
	msg->msg_namelen = sizeof( l2tp_session->tunnel_addr.addr );
	
	msg->msg_control = NULL;
	msg->msg_controllen = 0;
	msg->msg_flags = 0;
	
	error = l2tp_sock->old_proto.sendmsg( l2tp_session->server_sock, msg, total_len + sizeof( struct pppol2tp_hdr ) );
	
	kfree( msg->msg_iov );
	kfree( msg );
	
end:
	EXIT_FUNCTION;
	return error;
}


/************************************************************************
 *
 * xmit function for internal use.
 *
 ***********************************************************************/
MYSTATIC int __pppol2tp_xmit(struct sock *sk, struct sk_buff *skb)
{
	int error = 0;
	struct pppol2tp_hdr hdr;
	struct msghdr *msg;
	struct pppol2tp_session *l2tp_session;
	struct pppol2tp_sock *l2tp_sock;
	
	ENTER_FUNCTION;
	
	if (sk->dead || !(sk->state & PPPOX_CONNECTED)) {
		error = -ENOTCONN;
		goto end;
	}

	/* FIXME: Check magic */
	SOCK_2_SESSION( sk, l2tp_session, error, -EBADF, end );
	SOCK_2_SERVER( l2tp_session->server_sock, l2tp_sock, error, -EBADF, end );

	/* Setup L2TP header */	
	hdr.flags  = htons(L2TP_HDR_VER | L2TP_HDRFLAG_O);
	hdr.tunnel = htons( l2tp_session->tunnel_addr.d_tunnel );
	hdr.session= htons( l2tp_session->tunnel_addr.d_session );
	hdr.offset = 0;

	/* The old code had a size limit here. It's not really an issue
	 * since the upper layers will fragment anyway if necessary */
	
	/* Unfortunatly there doesn't appear to be a way for us to pass an
	 * skb to the UDP layer, we have to pretend to be sending ordinary
	 * data and use sendmsg */
	
	msg = kmalloc( sizeof( struct msghdr ), GFP_KERNEL );
	
	msg->msg_iov = kmalloc( 2 * sizeof( struct iovec ), GFP_KERNEL );
	msg->msg_iov[0].iov_base = &hdr;
	msg->msg_iov[0].iov_len  = sizeof( hdr );
	msg->msg_iov[1].iov_base = skb->data;
	msg->msg_iov[1].iov_len  = skb->data_len;
	msg->msg_iovlen = 2;
	
	/* If the user calls sendto() that's just too bad */
	msg->msg_name = &l2tp_session->tunnel_addr.addr;
	msg->msg_namelen = sizeof( l2tp_session->tunnel_addr.addr );
	
	msg->msg_control = NULL;
	msg->msg_controllen = 0;
	msg->msg_flags = 0;
	
	error = l2tp_sock->old_proto.sendmsg( l2tp_session->server_sock, msg, skb->data_len + sizeof( hdr ) );
	
	kfree( msg->msg_iov );
	kfree( msg );
	
end:
	EXIT_FUNCTION;
	return error;
}


/************************************************************************
 *
 * xmit function called by generic PPP driver
 * sends PPP frame over PPPoE socket
 *
 ***********************************************************************/
MYSTATIC int pppol2tp_xmit(struct ppp_channel *chan, struct sk_buff *skb)
{
	struct sock *sk = (struct sock *) chan->private;
	return __pppol2tp_xmit(sk, skb);
}

/* FIXME: Need to check for overflows, I don't quite remember how this works anymore */
int pppol2tp_proc_info(char *buffer, char **start, off_t offset, int length)
{
	struct pppol2tp_session *po;
	struct pppol2tp_sock *server_sock = pppol2tp_socket_list;
	int len = 0;
	off_t pos = 0;
	off_t begin = 0;
	int size;
	int i;
	
	ENTER_FUNCTION;

	len += sprintf(buffer,
		       "sock *  :l_sock *:userdata   Address\n" );
	pos = len;

	write_lock_bh(&pppol2tp_hash_lock);

	while( server_sock )
	{
		size = sprintf(buffer + len, "%p:%p:%p   %08X:%d %c MAGIC %s\n", 
				server_sock->sock,
				server_sock, 
				server_sock->sock->user_data,
				htonl( server_sock->sock->saddr ), 
				htons( server_sock->sock->sport ),
				(server_sock == server_sock->sock->user_data) ?'Y':'N',
				(server_sock->magic == L2TP_MAGIC)?"OK":"BAD" );

		len += size;
		pos += size;

		size = sprintf(buffer + len,
		       "  sock *  :session*:userdata   Address :Port    SAddrT:S  -> DAddrT:S    S\n");

		len += size;
		pos += size;
		for (i = 0; i < PPPOL2TP_HASH_SIZE; i++) {
			po = server_sock->hash[i];
			while (po) {
				size = sprintf(buffer + len,
					       "  %p:%p:%p   %08X:%-5d   %04X/%04X -> %04X/%04X   %d %c MAGIC %s\n",
					       po->sock,
					       po,
					       po->sock->user_data,
					       htonl( po->tunnel_addr.addr.sin_addr.s_addr ),
					       htons( po->tunnel_addr.addr.sin_port ),
					       po->tunnel_addr.s_tunnel, po->tunnel_addr.s_session,
					       po->tunnel_addr.d_tunnel, po->tunnel_addr.d_session,
					       po->sock->state,
					       (po == po->sock->user_data) ?'Y':'N',
					       (po->magic == L2TP_SESSION_MAGIC)?"OK":"BAD" );
				len += size;
				pos += size;
				if (pos < offset) {
					len = 0;
					begin = pos;
				}

				if (pos > offset + length)
					break;

				po = po->next;
			}

			if (po)
				break;
	  	}
	  	server_sock = server_sock->next;
	}
	write_unlock_bh(&pppol2tp_hash_lock);

  	*start = buffer + (offset - begin);
  	len -= (offset - begin);
  	if (len > length)
  		len = length;
	if (len < 0)
		len = 0;

	EXIT_FUNCTION;
  	return len;
}


MYSTATIC struct proto_ops pppol2tp_ops = {
    family:		AF_PPPOX,
    release:		pppol2tp_release,
    bind:		sock_no_bind,
    connect:		pppol2tp_connect,
    socketpair:		sock_no_socketpair,
    accept:		sock_no_accept,
    getname:		pppol2tp_getname,
    poll:		datagram_poll,
    ioctl:		pppol2tp_ioctl,
    listen:		sock_no_listen,
    shutdown:		sock_no_shutdown,
    setsockopt:		sock_no_setsockopt,
    getsockopt:		sock_no_getsockopt,
    sendmsg:		pppol2tp_sendmsg,
    recvmsg:		pppol2tp_recvmsg,
    mmap:		sock_no_mmap
};

struct pppox_proto pppol2tp_proto = {
    create:	pppol2tp_create,
    ioctl:	pppol2tp_ioctl
};


int __init pppol2tp_init(void)
{
 	int err = register_pppox_proto(PX_PROTO_L2TP, &pppol2tp_proto);

	if (err == 0) {
		proc_net_create("pppol2tp", 0, pppol2tp_proc_info);
	}
	return err;
}

void __exit pppol2tp_exit(void)
{
	unregister_pppox_proto(PX_PROTO_L2TP);
	proc_net_remove("pppol2tp");
}

module_init(pppol2tp_init);
module_exit(pppol2tp_exit);

MODULE_AUTHOR("Martijn van Oosterhout <kleptog@svana.org>");
MODULE_DESCRIPTION("PPP over L2TP over UDP");
MODULE_LICENSE("GPL");
